'''
This place-holder module makes the extensions directory
into a Python "package", so that external user-specific
modules can act as umbrella modules, and, for example:

import rst2pdf.extensions.preprocess_r2p

to bring in the preprocess extension.
'''
